package com.vickingur;

import java.io.IOException;
import java.io.InputStream;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.graphics.Paint.Cap;
import android.graphics.Point;
import android.graphics.Rect;
import android.util.Log;
import android.view.SurfaceHolder;

import com.vickingur.model.GameProcessorFactory;
import com.vickingur.model.IGameProcessor;
import com.vickingur.model.ShapeFactory;
import com.vickingur.model.shapes.IShape;
import com.vickingur.util.Constants;
import com.vickingur.util.DrawUtil;

public class PainterThread extends Thread {

    public static final int READY = 1;

    public static final int SLEEP = 0;

    private Bitmap bitmap;

    private Paint brush;

    private float brushSize;

    // private Canvas canvas;

    private int canvasBgColor;

    private SurfaceHolder holder;

    private boolean isActive;

    private int status;

    private Point topLeft = new Point(0, 0);

    private int alive = 0;

    private Context context;

    private IGameProcessor processor;

    public PainterThread(SurfaceHolder surfaceHolder, Context context) {
        super("PainterThread");
        processor = GameProcessorFactory.getInstance(Constants.IS_NATIVE_IMPL);
        holder = surfaceHolder;
        this.context = context;
        brushSize = 2;
        brush = new Paint();
        brush.setAntiAlias(true);
        brush.setColor(Color.rgb(0, 0, 0));
        brush.setStrokeWidth(brushSize);
        brush.setStrokeCap(Cap.ROUND);
        canvasBgColor = Color.WHITE;
    }

    public void activate() {
        status = PainterThread.READY;
    }

    public void freeze() {
        status = PainterThread.SLEEP;
    }

    public boolean isFreeze() {
        return (status == PainterThread.SLEEP);
    }

    public boolean isReady() {
        return (status == PainterThread.READY);
    }

    public boolean isRunning() {
        return isActive;
    }

    public void off() {
        isActive = false;
    }

    public void on() {
        isActive = true;
    }

    @Override
    public void run() {
        Log.v(getName(), "waiting for bitmap");
        waitForBitmap();
        Log.v(getName(), "finished waiting for bitmap");
        processor.createField(bitmap.getHeight() / Constants.SIZE_DIVIDER,
                bitmap.getWidth() / Constants.SIZE_DIVIDER);

        while (isRunning()) {
            if (GameProcessorFactory.getInstance(Constants.IS_NATIVE_IMPL)
                    .isNeedToCleanUpAndStop()) {
                off();
            }
            if (isFreeze()) {
                try {
                    Thread.sleep(Constants.SLEEP_TIME);
                } catch (InterruptedException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
                continue;
            }
            long endTime = System.nanoTime();
            endTime += Constants.DRAWING_PERIOD * Constants.MEGA;
            Canvas c = null;
            try {
                c = holder.lockCanvas();
                synchronized (holder) {
                    if (status == PainterThread.READY) {
                        long startDraw = System.nanoTime();
                        Log.v(getName(), "drawing: " + bitmap);
                        c.drawBitmap(bitmap, topLeft.x, topLeft.y, null);
                        DrawUtil.drawField(processor.getCells(), c, brush,
                                topLeft, processor.getRows(),
                                processor.getColumns());
                        Paint p = new Paint();
                        p.setColor(Color.DKGRAY);
                        p.setAntiAlias(true);
                        c.drawRect(new Rect(0, 0,
                                topLeft.x + bitmap.getWidth(), topLeft.y), p);
                        p.setColor(Color.LTGRAY);
                        int rows = processor.getRows();
                        int cols = processor.getColumns();
                        c.drawText("gen: " + processor.getGenerationCount()
                                + "  alive cells: " + alive + "  rows: " + rows
                                + "  cols: " + cols, topLeft.x,
                                3 * topLeft.y / 4, p);
                        long endDraw = System.nanoTime();
                        Log.v(getName(), "drawing time: "
                                + ((endDraw - startDraw) / Constants.MEGA)
                                + " ms");
                    }
                }
            } finally {
                if (c != null) {
                    holder.unlockCanvasAndPost(c);
                }
                long startTick = System.nanoTime();
                alive = processor.tick();
                long endTick = System.nanoTime();
                Log.v(getName(), "game tick took "
                        + ((endTick - startTick) / Constants.MEGA) + " ms");
                long actualEndTime = System.nanoTime();
                Log.v(getName(), "endTime: " + endTime + " actualEndTime: "
                        + actualEndTime);
                int delta = (int) ((endTime - actualEndTime) > 0 ? (endTime - actualEndTime)
                        : 0);
                try {
                    Log.v(getName(), "Sleeping for " + delta + " nanoseconds");
                    Thread.sleep(delta / Constants.MEGA);
                } catch (InterruptedException e) {
                }
            }
        }
        Log.v(getName(), "Exiting app");
        System.exit(0);
    }

    private void putTestShapes() {
        // IShape glider = ShapeFactory.createNewShape(IShape.Glider, null);
        // glider.build(Constants.TEST_START_POINT_GLIDER);
        // GameProcessor.getInstance().putShape(glider);
        //
        // IShape penta = ShapeFactory.createNewShape(IShape.Pentadecathlon,
        // null);
        // penta.build(Constants.TEST_START_POINT_PENTA);
        // GameProcessor.getInstance().putShape(penta);
        //
        // IShape tipper = ShapeFactory.createNewShape(IShape.Tipper, null);
        // tipper.build(Constants.TEST_START_POINT_TIPPER);
        // GameProcessor.getInstance().putShape(tipper);
        //
        // IShape eight = ShapeFactory.createNewShape(IShape.Eight, null);
        // eight.build(Constants.TEST_START_POINT_EIGHT);
        // GameProcessor.getInstance().putShape(eight);
        //
        // IShape boat = ShapeFactory.createNewShape(IShape.Boat, null);
        // boat.build(Constants.TEST_START_POINT_BOAT);
        // GameProcessor.getInstance().putShape(boat);
        //
        // IShape bheptomino = ShapeFactory.createNewShape(IShape.BHeptomino,
        // null);
        // bheptomino.build(Constants.TEST_START_POINT_BHEPTOMINO);
        // GameProcessor.getInstance().putShape(bheptomino);

        InputStream is = null;
        try {
            is = context.getAssets().open(
                    Constants.IMPORT_SHAPES_DIR + "/"
                            + Constants.IMPORT_TEST_NAME);
        } catch (IOException e) {
            Log.e(getName(), "IOException for file: "
                    + Constants.IMPORT_TEST_NAME, e);
        }
        IShape imported = ShapeFactory.createNewShape(IShape.ImportedShape, is);
        imported.build(Constants.TEST_START_POINT_IMPORTED);
        processor.addShapeToQ(imported);
    }

    public void setBitmap(Bitmap bitmap, boolean clear) {
        this.bitmap = bitmap;
        if (clear) {
            bitmap.eraseColor(canvasBgColor);
        }

        // this.canvas = new Canvas(this.bitmap);
    }

    private void waitForBitmap() {
        while (bitmap == null) {
            try {
                Thread.sleep(Constants.SLEEP_TIME);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }
    }

    public void setTopLeft(Point topLeft) {
        this.topLeft = topLeft;
    }

}
