package com.vickingur.model;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;

import android.graphics.Color;
import android.graphics.Point;

import com.vickingur.model.shapes.Cell;
import com.vickingur.model.shapes.IShape;

public class GameProcessor implements IGameProcessor {

    private static final int overpopulationNumber = 4;
    private static final int starvationNumber = 2;
    private Queue<IShape> shapeQ = new LinkedList<IShape>();

    private Map<Point, Integer> colors = new HashMap<Point, Integer>();

    private Field field;
    private boolean paused = false;
    private int alive = 0;
    private int ticks = 0;
    private boolean needToCleanUpAndStop;

    private CellAction analyzeCell(Point point) {
        CellAction action = CellAction.Die;
        int numAlive = calcAliveNeighbours(point);
        if (analyzeCellState(field.getCells()[point.x * field.getColumns()
                + point.y]) == CellState.Alive) {
            if ((numAlive < starvationNumber)
                    || (numAlive >= overpopulationNumber)) {
                action = CellAction.Die;
            } else {
                action = CellAction.Live;
                colors.put(point, field.getCells()[point.x * field.getColumns()
                        + point.y].getColor());
            }
        } else {
            if (numAlive == 3) {
                action = CellAction.Born;
                int color = getDominatingNeighbouringColor(point);
                colors.put(point, color);
            }
        }
        return action;
    }

    private CellState analyzeCellState(Cell cell) {
        CellState state = CellState.Alive;
        if (cell.getColor() == Color.WHITE) {
            state = CellState.Dead;
        }
        return state;
    }

    private int calcAliveNeighbours(Point point) {
        int num = 0;

        ArrayList<Point> toAnalyze = new ArrayList<Point>();
        toAnalyze.add(new Point(point.x - 1, point.y));
        toAnalyze.add(new Point(point.x + 1, point.y));
        toAnalyze.add(new Point(point.x - 1, point.y - 1));
        toAnalyze.add(new Point(point.x + 1, point.y - 1));
        toAnalyze.add(new Point(point.x - 1, point.y + 1));
        toAnalyze.add(new Point(point.x + 1, point.y + 1));
        toAnalyze.add(new Point(point.x, point.y - 1));
        toAnalyze.add(new Point(point.x, point.y + 1));
        for (Point p : toAnalyze) {
            Point valid = Field.validatePoint(p, field.getRows(),
                    field.getColumns());
            if (analyzeCellState(field.getCells()[valid.x * field.getColumns()
                    + valid.y]) == CellState.Alive) {
                num++;
            }
        }
        alive = num;
        return num;
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.vickingur.model.IGameProcessor#createField(int, int)
     */
    @Override
    public void createField(int rows, int columns) {
        field = new Field(rows, columns);
    }

    private int getDominatingNeighbouringColor(Point point) {
        int domColor = -1;
        Map<Integer, Integer> colors = new HashMap<Integer, Integer>();
        ArrayList<Point> toAnalyze = new ArrayList<Point>();
        toAnalyze.add(new Point(point.x - 1, point.y));
        toAnalyze.add(new Point(point.x + 1, point.y));
        toAnalyze.add(new Point(point.x - 1, point.y - 1));
        toAnalyze.add(new Point(point.x + 1, point.y - 1));
        toAnalyze.add(new Point(point.x - 1, point.y + 1));
        toAnalyze.add(new Point(point.x + 1, point.y + 1));
        toAnalyze.add(new Point(point.x, point.y - 1));
        toAnalyze.add(new Point(point.x, point.y + 1));
        for (Point p : toAnalyze) {
            Point valid = Field.validatePoint(p, field.getRows(),
                    field.getColumns());
            int color = field.getCells()[valid.x * field.getColumns() + valid.y]
                    .getColor();
            if (color != Color.WHITE) {
                if (colors.containsKey(color)) {
                    int val = colors.get(color);
                    colors.put(color, val + 1);
                } else {
                    colors.put(color, 1);
                }
            }
        }
        int max = -1;
        for (Map.Entry<Integer, Integer> entry : colors.entrySet()) {
            if (max < entry.getValue()) {
                max = entry.getValue();
                domColor = entry.getKey();
            }
        }
        return domColor;
    }

    public Field getField() {
        return field;
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.vickingur.model.IGameProcessor#tick()
     */
    @Override
    public synchronized int tick() {
        while (!shapeQ.isEmpty()) {
            IShape shape = shapeQ.poll();
            putShape(shape);
        }
        if (paused)
            return alive;
        int count = 0;
        colors.clear();
        int rows = field.getRows();
        int columns = field.getColumns();
        CellAction[][] ca = new CellAction[rows][columns];
        for (int i = 0; i < rows; i++) {
            for (int j = 0; j < columns; j++) {
                if (analyzeCell(new Point(i, j)) == CellAction.Die) {
                    ca[i][j] = CellAction.Die;
                } else {
                    ca[i][j] = CellAction.Live;
                }
                // ca[i][j] = analyzeCell(new Point(i,j));
            }
        }
        for (int i = 0; i < rows; i++) {
            for (int j = 0; j < columns; j++) {
                if (ca[i][j] == CellAction.Live) {
                    field.getCells()[i * columns + j] = new Cell(
                            colors.get(new Point(i, j)), new Point(i, j));
                    count++;
                } else {
                    field.getCells()[i * columns + j] = new Cell(Color.WHITE,
                            new Point(i, j));
                }
            }
        }
        alive = count;
        ticks++;
        return count;
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.vickingur.model.IGameProcessor#pauseGame()
     */
    @Override
    public void pauseGame() {
        paused = true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.vickingur.model.IGameProcessor#resumeGame()
     */
    @Override
    public void resumeGame() {
        paused = false;
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.vickingur.model.IGameProcessor#clear()
     */
    @Override
    public synchronized void clear() {
        ticks = 0;
        field.clear();
        resumeGame();
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.vickingur.model.IGameProcessor#getGenerationCount()
     */
    @Override
    public int getGenerationCount() {
        return ticks;
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.vickingur.model.IGameProcessor#switchState()
     */
    @Override
    public boolean switchState() {
        paused = !paused;
        return paused;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * com.vickingur.model.IGameProcessor#putShape(com.vickingur.model.shapes
     * .IShape)
     */
    @Override
    public synchronized void putShape(IShape shape) {
        field.putShape(shape);
    }

    @Override
    public Cell[] getCells() {
        return field.getCells();
    }

    @Override
    public int getRows() {
        return field.getRows();
    }

    @Override
    public int getColumns() {
        return field.getColumns();
    }

    @Override
    public void addShapeToQ(IShape shape) {
        shapeQ.add(shape);
    }

    @Override
    public void requestCleanUp() {
        needToCleanUpAndStop = true;
    }

    @Override
    public boolean isNeedToCleanUpAndStop() {
        return needToCleanUpAndStop;
    }

}
